/* 
** pipe_server/main.c 
** http://graphics.stanford.edu/software/wiregl
**
** Copyright 2000
** The Board of Trustees of The Leland Stanford Junior University.
** All rights reserved.
**
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or
** modify this software, provided that the above copyright notice and
** this permission notice appear in all copies of this software.  No
** part of this software or any derivatives thereof may be used in
** graphics systems for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind,
** express, implied or otherwise, including without limitation, any
** warranty of merchantability or fitness for a particular purpose.
*/

#ifdef WINDOWS
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif
#include <GL/gl.h>


#include "wiregl/include/wiregl_pipe.h"
#include "wiregl/include/wiregl_protocol.h"
#include "wiregl/include/wiregl_util.h"
#include "wiregl/include/wiregl_common.h"

#include "wiregl/glcontext/glcontext.h"

WireGLPipe wiregl_pipe;
static int fullscreen_mode;

/* XXX HACK: This should go into a header file, but which one? */
extern GLcurrent current;

static void __parseTrace( char *prog, char **file )
{
	wiregl_pipe.fname = *file;
	fprintf( stderr, "%s: trace files not currently supported\n",
		prog );
	exit( 1 );
}

static void __parsePort( char *prog, char **port )
{
	int x;

	x = atoi( *port );
	if ( x <= 0 || x > 0xffff )
	{
		fprintf( stderr, "%s: bad port number: %s\n", 
			prog, *port );
		exit( 1 );
	}
	wiregl_pipe.port = (unsigned short) x;
}

static void __parseFullscreen( char *prog, char **args )
{
	(void)prog;
	(void)args;

	fullscreen_mode = 1;
}

typedef struct {
	char *flag;
	void (*fn)(char *prog, char **args);
	int numParameters;
} ParseParameterT;

ParseParameterT parse_structs[] = {
	{"-p", __parsePort, 1},
	{"-t", __parseTrace, 1},
	{"-f", __parseFullscreen, 0},
	{NULL, NULL, 0}
};

static void __setDefaults( void )
{
	wiregl_pipe.port = WIREGL_TCPIP_DEFAULT_PORT;
	wiregl_pipe.window_x = 0;
	wiregl_pipe.window_y = 0;
	wiregl_pipe.window_width = 1024;
	wiregl_pipe.window_height = 768;
	wiregl_pipe.fname = NULL;
	wiregl_pipe.swap_barrier = 0;
	wiregl_pipe.barrier_conn = NULL;
	fullscreen_mode = 0;
}

static void __parseArgs( int argc, char **argv )
{
	int i,j;

	__setDefaults();

	for (i = 1 ; i < argc ; i++)
	{
		for (j = 0 ; parse_structs[j].flag ; j++)
		{
			if (!strcmp( argv[i], parse_structs[j].flag ))
			{
				if( argc - i - 1 < parse_structs[j].numParameters )
				{
					fprintf( stderr, "%s: %s expects %d arguments", argv[0], parse_structs[j].flag,
						parse_structs[j].numParameters );
					exit( 1 );
				}
				
				parse_structs[j].fn( argv[0], &argv[i + 1] );
				i+=parse_structs[j].numParameters;
				break;
			}
		}

		if(!parse_structs[j].flag)
		{
			fprintf( stderr, "%s: bad argument \"%s\"\n", argv[0], argv[i] );
			exit( 1 );
		}
	}
}

int main( int argc, char *argv[] )
{
	GLconfig *c;

	wireGLInitInstrumentation( TO_SCREEN );
	wireGLInitOS( );
	wireGLNetInit( wireGLPipeserverReceive, NULL );
	wireGLInitWarnings( getenv( "WIREGL_WARN_LEVEL" ),
						( getenv( "WIREGL_GUI_ERROR" ) != NULL ) );

	__parseArgs( argc, argv );

	c = __glconfig_CreateDefaultConfig( );
	c->mode = (GLenum) GLCONFIG_TRACK;
	c->current = &current;
	__glInit( c );

	wireGLInitEvent( &work_pending );
	wireGLInitLock( &new_client_lock );
	wireGLInitSemaphore( &window_sema, 0 );

	wireGLStartConnectionThread( );
	wireGLCreateWindow( fullscreen_mode );
	//wireGLChangeThreadPriority( WIREGL_HIGH_PRIORITY );
	wireGLRenderRemoteStreams( );

	return 0;
}
