/* 
** pipe_server/stream.c 
** http://graphics.stanford.edu/software/wiregl
**
** Copyright 2000
** The Board of Trustees of The Leland Stanford Junior University.
** All rights reserved.
**
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or
** modify this software, provided that the above copyright notice and
** this permission notice appear in all copies of this software.  No
** part of this software or any derivatives thereof may be used in
** graphics systems for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind,
** express, implied or otherwise, including without limitation, any
** warranty of merchantability or fitness for a particular purpose.
*/

#ifdef WINDOWS
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif
#include <GL/gl.h>

#include <stdio.h>
#include <math.h>
#include "wiregl/include/wiregl_opcodes.h"
#include "wiregl/include/wiregl_protocol.h"
#include "wiregl/include/wiregl_pipe.h"
#include "wiregl/include/wiregl_util.h"
#include "wiregl/glcontext/glcontext.h"

#define FLOW_CONTROL 1
#define MAX_OUTSTANDING_BYTES ( 1 << 20 )

unsigned char *unpacker_data_ptr = NULL;
WireGLConnection *current_conn = NULL;
WireGLReceiveBuffer *receive_buffers = NULL;
WireGLEvent work_pending;
WireGLLock  new_client_lock;

static WireGLHashTable  *work_queues = NULL;
static WireGLWorkQueue  *work_ring   = NULL;

static struct {
	int num;
	int max;
	WireGLWorkQueue **queues;
} new_clients = { 0, 0, NULL };

WireGLWorkQueue *__currentWorkQueue( void ) { return work_ring; }

static WireGLWorkQueue *CreateWorkQueue( WireGLConnection *conn, WireGLConnection *peer_conn )
{
	WireGLWorkQueue *q;

	q = (WireGLWorkQueue *) wireGLAlloc( sizeof( *q ) );
	wireGLInitLock( &q->lock );
	q->context = __glCreateContext();
	q->total_length = 0;
	q->head = q->tail = NULL;
	q->blocked = 0;
	q->delete_when_done = 0;
	q->conn = conn;
	q->peer_conn = peer_conn;
	q->multi.length = 0;
	q->multi.max    = 0;
	q->multi.buffer = NULL;
	q->next = q->prev = NULL;

	return q;
}

static void DestroyWorkQueue( WireGLWorkQueue *q )
{
	wireGLAssert( q->head == NULL && q->tail == NULL );
	wireGLAssert( q->multi.buffer == NULL );
	wireGLDestroyLock( &q->lock );
	__glDestroyContext( q->context );
	wireGLFree( q );
}

void wireGLPipeserverConnect( WireGLConnection *conn, WireGLConnection *peer )
{
	WireGLWorkQueue *q;

	wireGLWarning( WIREGL_WARN_NOTICE, "new client, sender_id=%u",
				   conn->sender_id );

	if ( work_queues == NULL )
	{
		/* our first time through... */
		work_queues = wireGLAllocHashtable();
		/* make sure the window is up first... */
		/*
		  wireGLAcquireSemaphore( &window_sema );
		*/
	}

	q = CreateWorkQueue( conn, peer );


	wireGLHashtableAdd( work_queues, conn->sender_id, q );

	wireGLAcquireLock( &new_client_lock );
	if ( new_clients.max == 0 )
	{
		new_clients.max = 1;
		new_clients.queues = (WireGLWorkQueue **) wireGLAlloc( new_clients.max * sizeof(new_clients.queues[0]) );
	}
	if ( new_clients.num == new_clients.max )
	{
		int n_bytes;
		new_clients.max <<= 1;
		n_bytes = new_clients.max * sizeof(new_clients.queues[0]);
		wireGLRealloc( (void **) &new_clients.queues, n_bytes );
	}
	new_clients.queues[ new_clients.num++ ] = q;
	wireGLReleaseLock( &new_client_lock );
}

void 
wireGLPipeserverDisconnect( unsigned int sender_id )
{
	WireGLWorkQueue *q;
	wireGLWarning( WIREGL_WARN_NOTICE, "disconnecting sender_id=%u",
				   sender_id );

	q = (WireGLWorkQueue *) wireGLHashtableSearch( work_queues, sender_id );
	if ( q ) 
	{
		/* don't call delete, it will free() the value! */
		/* wireGLHashtableDelete( work_queues, sender_id ); */
		q->delete_when_done = 1;
	}
	else
	{
		wireGLWarning( WIREGL_WARN_CRITICAL, "wireGLDisconnect: sender_id=%u,"
					   " no such queue?", sender_id );
	}
}

static WireGLReceiveBuffer *
AllocReceiveBuffer( void )
{
    WireGLReceiveBuffer *recv_buf;

    if ( !receive_buffers )
    {
        recv_buf = (WireGLReceiveBuffer *) wireGLAlloc( sizeof(*recv_buf) );
    }
    else
    {
        recv_buf = receive_buffers;
        receive_buffers = receive_buffers->next;
    }

    recv_buf->next = NULL;
    return recv_buf;
}

static void
FreeReceiveBuffer( WireGLReceiveBuffer *recv_buf )
{
    recv_buf->next = receive_buffers;
    receive_buffers = recv_buf;
}

static void 
PutReceiveBufferOnQueue( WireGLWorkQueue *q, WireGLReceiveBuffer *buf )
{
	wireGLAcquireLock( &q->lock );

	buf->next = NULL;

	if ( q->tail )
		q->tail->next = buf;
	else
		q->head = buf;
	q->tail = buf;

#if FLOW_CONTROL
	if ( q->total_length < MAX_OUTSTANDING_BYTES )
	{
		q->total_length += buf->length;
		if ( q->total_length >= MAX_OUTSTANDING_BYTES )
		{
#if 0
			wireGLWarning( WIREGL_WARN_DEBUG, "FlowControl: sender_id=%d "
						   "enable=%d", q->conn->sender_id, 0 );
#endif
			q->conn->FlowControl( q->conn, 0 );
		}
	}
	else
	{
		q->total_length += buf->length;
	}

	/*
	  wireGLWarning( WIREGL_WARN_DEBUG, "q->total_length=%d",
	  q->total_length );
	*/
#endif

	wireGLReleaseLock( &q->lock );
}

void 
wireGLPipeserverReceive( void *buf, unsigned int len, WireGLFreeFunc freeFunc )
{
	WireGLMessage   *msg;
	WireGLWorkQueue *q;

	wireGLAssert( work_queues );
	
	msg = (WireGLMessage *) buf;
	switch ( msg->type )
	{
#if 0
	  case WIREGL_MESSAGE_MULTI_BODY:
	  case WIREGL_MESSAGE_MULTI_TAIL:
		q = (WireGLWorkQueue *) 
			wireGLHashtableSearch( work_queues, msg->multi.senderId );
		wireGLAssert( q );
		if ( len + q->multi.length > q->multi.max )
		{
			if ( q->multi.max == 0 ) 
			{
				q->multi.length = 0;
				q->multi.max    = 8192;
			}
			while ( len + q->multi.length > q->multi.max )
			{
				q->multi.max <<= 1;
			}
			wireGLRealloc( &q->multi.buffer, q->multi.max);
		}
		{
			unsigned char *src, *dst;
			dst = (unsigned char *) q->multi.buffer + q->multi.length;
			src = (unsigned char *) buf + sizeof(WireGLMessageMulti);
			len -= sizeof(WireGLMessageMulti);
			memcpy( dst, src, len );
			q->multi.length += len;
		}

		freeFunc( buf );

		if ( msg->type == WIREGL_MESSAGE_MULTI_TAIL )
		{
			WireGLReceiveBuffer *multi = AllocReceiveBuffer( );

			multi->freeFunc = wireGLFree;
			multi->length   = q->multi.length;
			multi->buffer   = q->multi.buffer;

			q->multi.length = 0;
			q->multi.max    = 0;
			q->multi.buffer = NULL;

			PutReceiveBufferOnQueue( q, multi );

			wireGLSignalEvent( &work_pending );

		}

		break;
#endif

	  case WIREGL_MESSAGE_OPCODES:
		{
			WireGLReceiveBuffer *recv_buf;

			q = (WireGLWorkQueue *) 
				wireGLHashtableSearch( work_queues, msg->opcodes.senderId );
			wireGLAssert( q && q->multi.buffer == NULL );

			recv_buf = AllocReceiveBuffer( );
			recv_buf->freeFunc = freeFunc;
			recv_buf->length   = len;
			recv_buf->buffer   = buf;

			PutReceiveBufferOnQueue( q, recv_buf );
		
			if (q->peer_conn) {
				/* Do some ring stuff */
			}

			wireGLSignalEvent( &work_pending );

			break;
		}

	  default:
		wireGLError( "wireGLPipeserverReceive: msg->type=0x%x?", msg->type );
		break;
	}
}

static WireGLWorkQueue *GetNextQueue( void )
{
	int i;

	for ( ; ; )
	{
		WireGLWorkQueue *q;

		if ( work_ring && work_ring->delete_when_done )
		{
			/* only delete this queue when all its buffers are empty
			 * (presumably we got the delete notification in order,
			 * and once we've emptied all the current buffers we won't
			 * get any more) */

			q = work_ring;
			if ( q->head == NULL )
			{
				if ( q == q->next )
				{
					work_ring = NULL;
				}
				else
				{
					work_ring     = q->next;
					q->prev->next = q->next;
					q->next->prev = q->prev;
				}

				/* additionally, if the queue is blocked then there is
				 * a pointer in a parallel api structure somewhere to
				 * it, so it will get unblocked eventually.  We should
				 * wait for it to get unblocked.  Unfortunately, if
				 * the client that will unblock this queue exits, then
				 * the semaphoreV/barrier/etc. will never arrive.
				 * Gargh.  Should we be reference couting these?
				 * Uggggly. */
				if ( !q->blocked )
				{
					wireGLWarning( WIREGL_WARN_NOTICE, "destroying q=0x%p "
								   "sender_id=%u", q, q->conn->sender_id );
					DestroyWorkQueue( q );
				}
				else
				{
					wireGLWarning( WIREGL_WARN_CRITICAL, "it's time to "
								   "release the resources held by "
								   "sender_id=%u, but it tagged as blocked, "
								   "which means there is an outstanding "
								   "pointer to them.", q->conn->sender_id );
				}
			}
		}

		/* place all the new clients on to the work ring */
		wireGLAcquireLock( &new_client_lock );

		for ( i = 0; i < new_clients.num; i++ )
		{
			q = new_clients.queues[i];
			if ( !work_ring )
			{
				work_ring = q;
				q->next = q;
				q->prev = q;
			}
			else
			{
				q->next = work_ring->next;
				work_ring->next->prev = q;

				q->prev = work_ring;
				work_ring->next = q;
			}
		}
		
		new_clients.num = 0;

		wireGLReleaseLock( &new_client_lock );

		/* now look for something to do */
		if ( work_ring )
		{
			int all_blocked = 1;
			WireGLWorkQueue *start = work_ring;
			while ( work_ring->next != start ) 
			{
				if ( !work_ring->blocked )
					all_blocked = 0;

				/* work on this guy if he isn't blocked and he has at
				 * least 1 buffer of work to do */
				if ( !work_ring->blocked && work_ring->head )
					return work_ring;

				work_ring = work_ring->next;
			}

			if ( !work_ring->blocked )
				all_blocked = 0;

			if ( !work_ring->blocked && work_ring->head )
				return work_ring;

			if ( all_blocked )
			{
				wireGLWarning( WIREGL_WARN_CRITICAL, "all the available "
							   "clients are blocked" );
			}
		}

#if 0
		wireGLWaitEvent( &work_pending );
#else
		wireGLNetRecv( );
		wireGLSleep( 0 );
#endif
	}

	/* UNREACHABLE */
	/* return NULL; */
}


static WireGLReceiveBuffer *GetNextBufferFromQueue( WireGLWorkQueue *q )
{
	WireGLReceiveBuffer *buf;

	wireGLAcquireLock( &q->lock );

	buf = q->head;
	if ( buf ) 
	{
		q->head = q->head->next;
		if ( !q->head )
			q->tail = NULL;

#ifdef FLOW_CONTROL
		if ( q->total_length >= MAX_OUTSTANDING_BYTES ) 
		{
			q->total_length -= buf->length;
			if ( q->total_length < MAX_OUTSTANDING_BYTES )
			{
#if 0
				wireGLWarning( WIREGL_WARN_DEBUG, "FlowControl: sender_id=%d "
							   "enable=%d", q->conn->sender_id, 1 );
#endif
				q->conn->FlowControl( q->conn, 1 );
			}
		}
		else
		{
			q->total_length -= buf->length;
		}
		wireGLAssert( ( q->total_length > 0 && q->head ) || 
					  ( q->total_length == 0 && !q->head ) );
#endif
	}

	wireGLReleaseLock( &q->lock );

	return buf;
}

void wireGLRenderRemoteStreams( void )
{
	for (;;)
	{
		int i;
		int switch_contexts = 0;
		WireGLReceiveBuffer *buf;

		WireGLWorkQueue *q = GetNextQueue( ); 

		__glMakeCurrent( q->context );

		buf = GetNextBufferFromQueue( q );
		while ( buf )
		{
			unsigned char *opcodes;
			WireGLMessageOpcodes *hdr = (WireGLMessageOpcodes *) buf->buffer;
			int num_opcodes = hdr->numOpcodes;
			unpacker_data_ptr = (unsigned char *) buf->buffer + sizeof(*hdr) + (( num_opcodes + 3 ) & ~0x3 );
			opcodes = unpacker_data_ptr - 1;
			current_conn = q->conn;

#if 0
			/* HACK HACK HACK */
			Sleep( 10 /* milliseconds */ );
#endif

			for (i = 0 ; i > -(num_opcodes) ; i--)
			{
				unsigned char op = opcodes[i];
#if 0
				printf ("%d/%d: ptr=0x%p ", -i+1, num_opcodes,
						unpacker_data_ptr );
				wireGLPrintFunc( (WireGLOpcode) op, 
								 (unsigned char *) unpacker_data_ptr );
				if ( op == WIREGL_SWAPBUFFERS_OPCODE )
					printf("--- swap ---------------------\n");
#endif
				wireGLAssert( op <= WIREGL_NOP_OPCODE );
				wiregl_decode_functions[op]( );
				wireGLAssert( unpacker_data_ptr <= (unsigned char *) buf->buffer + buf->length );
			}

			/* we've stopped rendering, lets grab the current state */
			__glExtractCurrentState();

			/* a round-about way of saying "if we are blocked, it had
			 * better be because the last opcode in the block (the one
			 * we'll test to set switch_contexts) caused it */
			wireGLAssert( ( q->blocked && switch_contexts ) || !q->blocked );

			buf->freeFunc( buf->buffer );
			FreeReceiveBuffer( buf );

			if ( switch_contexts )
				break;

			buf = GetNextBufferFromQueue( q );
		}

		work_ring = work_ring->next;
	}
}

