/*
** progs/wgl/wgl.c
** Written by Mathew Eldridge 12-Oct-99
** support for running applications under wireGL
** Generated by opengl_stub/windows_exports.py
** http://graphics.stanford.edu/software/wiregl
**
** Copyright 2000 
** The Board of Trustees of The Leland Stanford Junior University.
** All rights reserved.
**
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or
** modify this software, provided that the above copyright notice and
** this permission notice appear in all copies of this software.  No
** part of this software or any derivatives thereof may be used in
** graphics systems for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind,
** express, implied or otherwise, including without limitation, any
** warranty of merchantability or fitness for a particular purpose.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <process.h>
#else
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <dirent.h>
#endif

#ifdef _WIN32

#define DEFAULT_TMP_DIR "C:\\"

static const char *libgl_names[] = {
	"opengl32.dll"
};

#else

#define DEFAULT_TMP_DIR "/tmp"

#if defined(IRIX) || defined(IRIX64)
#define SYSTEM_LIB_DIR  "/usr/lib32"
#else
#define SYSTEM_LIB_DIR  "/usr/lib"
#endif

static const char *libgl_names[] = {
	"libGL.so"
};

#endif


static const char *progname   = "wgl";
static const char *wiregl_lib = NULL;
static int         verbose    = 0;

#ifdef _WIN32

static const char *
error_string( int err )
{
    char *buf;

    FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                   FORMAT_MESSAGE_FROM_SYSTEM, NULL, err,
                   MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
                   (LPTSTR) &buf, 0, NULL );
    return buf;
}

#endif

static void
debug( const char *format, ... )
{
	if ( verbose ) {
		va_list args;
		fprintf( stderr, "%s: ", progname );
		va_start( args, format );
		vfprintf( stderr, format, args );
		va_end( args );
	}
}

static void
error( const char *format, ... )
{
	va_list args;
	fprintf( stderr, "%s: ", progname );
	va_start( args, format );
	vfprintf( stderr, format, args );
	va_end( args );
}

static void
sys_error( const char *format, ... )
{
#ifdef _WIN32
	const char *msg = error_string( GetLastError( ) );
#else
	const char *msg = strerror( errno );
#endif

	va_list args;
	fprintf( stderr, "%s: ", progname );
	va_start( args, format );
	vfprintf( stderr, format, args );
	va_end( args );
	fprintf( stderr, ": %s", msg );
#ifndef _WIN32
	fputc( '\n', stderr );
#endif	
}

static void
fatal( const char *format, ... )
{
	va_list args;
	fprintf( stderr, "%s: ", progname );
	va_start( args, format );
	vfprintf( stderr, format, args );
	va_end( args );
	exit( 1 );
}

static void
sys_fatal( const char *format, ... )
{
#ifdef _WIN32
	const char *msg = error_string( GetLastError( ) );
#else
	const char *msg = strerror( errno );
#endif

	va_list args;
	fprintf( stderr, "%s: ", progname );
	va_start( args, format );
	vfprintf( stderr, format, args );
	va_end( args );
	fprintf( stderr, ": %s", msg );
#ifndef _WIN32
	fputc( '\n', stderr );
#endif	
	exit( 1 );
}

static const char *
basename( const char *path )
{
    char *last;
    last = strrchr( path, '/' );
    if ( !last )
        return path;
    else
        return last + 1;
}

static void *
xmalloc( size_t nbytes )
{
    void *ptr = malloc( nbytes );
    if ( ptr == NULL )
        fatal( "malloc( %d bytes ) failed!\n", nbytes );

    return ptr;
}

// So renamed to avoid conflicting with linux header files

static void
wgl_setenv( const char *var, const char *value )
{
    unsigned long len;
    char *buf;

    len = strlen(var) + 1 + strlen(value) + 1;
    buf = (char *) malloc( len );
    sprintf( buf, "%s=%s", var, value );
    putenv( buf );

	debug( "%s\n", buf );

    /* don't free the buf, the string is *part* of the environment,
     * and can't be reclaimed */
}

typedef struct List {
    char           *name;
    struct List *next;
} List;

static List *temp_files = NULL;
static List *temp_dirs  = NULL;

static void
add_to_list( List **list, const char *name )
{
	List *node = (List *) xmalloc( sizeof(*node) );

	node->name = (char *) xmalloc( strlen(name) + 1 );
	strcpy( node->name, name );
	node->next = *list;
	*list = node;
}

static void
add_file_to_temp_list( const char *filename )
{
	add_to_list( &temp_files, filename );
}

static void
delete_temp_files( void )
{
	List *list;

#ifdef _WIN32
	/* we seem to be in a race with a file lock, see if this fixes the
	 * problem */
	Sleep( 100 /* milliseconds */ );
#endif

	for ( list = temp_files; list; list = list->next ) {
#ifdef _WIN32
        if ( !DeleteFile( list->name ) )
            sys_error( "DeleteFile \"%s\"", list->name );
#else
		if ( unlink( list->name ) )
			sys_error( "unlink \"%s\"", list->name );
#endif
	}
}

static void
add_dir_to_temp_list( const char *dirname )
{
	add_to_list( &temp_dirs, dirname );
}

static void
delete_temp_dirs( void )
{
	List *list;
	for ( list = temp_dirs; list; list = list->next ) {
#ifdef _WIN32
		if ( !RemoveDirectory( list->name ) )
			sys_error( "RemoveDirectory \"%s\"", list->name );
#else
		if ( rmdir( list->name ) )
			sys_error( "rmdir \"%s\"", list->name );
#endif
	}
}

#ifdef _WIN32

char *
find_file_on_path( const char *name, char **tail_ptr )
{
    int   len;
    char *path, *tail;

    len = SearchPath( NULL, name, ".exe", 0, NULL, &tail );
    if ( len == 0 )
        sys_fatal( "\"%s\"", name );

	len+=2;
    path = (char *) xmalloc( len + 12 );

    if ( !SearchPath( NULL, name, ".exe", len, path, &tail ) )
        sys_fatal( "\"%s\"", name );

	if ( tail_ptr )
		*tail_ptr = tail;

	return path;
}

void
make_tmpdir( char *retval )
{
    char *name;

    name = _tempnam( DEFAULT_TMP_DIR, "wgl" );
    if ( name == NULL )
        fatal( "cannot create a unique directory name\n" );

	debug( "tmpdir=\"%s\"\n", name );

    if ( !CreateDirectory( name, NULL ) )
		sys_fatal( "CreateDirectory \"%s\"", name );

	strcpy( retval, name );
}

void
copy_file( const char *dst_filename, const char *src_filename )
{
	debug( "copying \"%s\" -> \"%s\"\n", src_filename, dst_filename );

    if ( !CopyFile( src_filename, dst_filename, 1 /* fail if exists */ ) )
        sys_fatal( "copy \"%s\" -> \"%s\"", src_filename, dst_filename );
}

static void
do_it( char *argv[] )
{
	char tmpdir[1024], argv0[1024], *tail;
    int i, status;

    make_tmpdir( tmpdir );
	add_dir_to_temp_list( tmpdir );

    argv[0] = find_file_on_path( argv[0], &tail );
	strcpy( argv0, tmpdir );
	strcat( argv0, "\\" );
	strcat( argv0, tail );

    copy_file( argv0, argv[0] );

	argv[0] = argv0;

	add_file_to_temp_list( argv0 );

	for ( i = 0; i < sizeof(libgl_names)/sizeof(libgl_names[0]); i++ ) {
		char name[1024];
		strcpy( name, tmpdir );
		strcat( name, "\\" );
		strcat( name, libgl_names[i] );
		copy_file( name, wiregl_lib );
		add_file_to_temp_list( name );
	}

    status = spawnv( _P_WAIT, argv[0], argv );

    if ( status == -1 )
        sys_fatal( "\"%s\"", argv[0] );
	else if ( status > 0 )
		error( "\"%s\": exited with status=%d\n", argv[0], status );

	delete_temp_files( );
	delete_temp_dirs( );

    exit( status ? 1 : 0 );
}

#else

static void
make_tmpdir( char *name )
{
	char *tmp;

	tmp = getenv( "TMP" );
	if ( !tmp )
		tmp = getenv( "TEMP" );
	if ( !tmp )
		tmp = DEFAULT_TMP_DIR;

	sprintf( name, "%s/wgl%d", tmp, (int) getpid( ) );

	debug( "tmpdir=\"%s\"\n", name );

	if ( mkdir( name, 0777 ) )
		sys_fatal( "mkdir \"%s\"", name );
}

static void
prefix_env_var( const char *prefix, const char *varname )
{
    char *val;

    val = getenv( varname );
    if ( val ) {
        unsigned long len = strlen( prefix ) + 1 + strlen( val ) + 1;
        char *buf = (char *) xmalloc( len );
        sprintf( buf, "%s:%s", prefix, val );
        wgl_setenv( varname, buf );
    } else {
        wgl_setenv( varname, prefix );
    }
}

static int
is_a_version_of( const char *basename, const char *libname )
{
	int len = strlen( basename );

	/* names must match */
	if ( strncmp( libname, basename, len ) ) return 0;

	libname += len;
	/* is this a version? */
	if ( *libname != '.' ) return 0;
		
	/* remainder of libname should be digits and periods */
	while ( *libname == '.' || isdigit(*libname) )
		libname++;

	/* did we make it to the end of the libname? */
	return ( *libname == '\0' );
}

static const char *
find_next_version_name( DIR *dir, const char *basename )
{
	struct dirent *entry;

	entry = readdir( dir );
	while ( entry ) {
		if ( is_a_version_of( basename, entry->d_name ) )
			return entry->d_name;
		entry = readdir( dir );
	}

	/* out of entries */
	return NULL;
}

static int
make_temp_link( const char *dir, const char *name, const char *target )
{
	char link_name[1024];

	strcpy( link_name, dir );
	strcat( link_name, "/" );
	strcat( link_name, name );

	if ( symlink( target, link_name ) ) {
		sys_error( "link \"%s\" -> \"%s\"", link_name, target );
		return 0;
	}

	add_file_to_temp_list( link_name );
	return 1;
}

static void
do_it( char *argv[] )
{
	pid_t pid;
	int status;
	char tmpdir[1024];
	int i;
	struct stat stat_buf;

	debug( "wiregl_lib=\"%s\"\n", wiregl_lib );

	if ( stat( wiregl_lib, &stat_buf ) )
		sys_fatal( "\"%s\"", wiregl_lib );

	if ( !S_ISREG(stat_buf.st_mode) )
		fatal( "\"%s\" isn't a regular file?\n", wiregl_lib );

	make_tmpdir( tmpdir );

	add_dir_to_temp_list( tmpdir );
	
	for ( i = 0; i < sizeof(libgl_names)/sizeof(libgl_names[0]); i++ ) {

		DIR *dir;

		make_temp_link( tmpdir, libgl_names[i], wiregl_lib );

		dir = opendir( SYSTEM_LIB_DIR );
		if ( dir ) {
			const char *version_name;
			version_name = find_next_version_name( dir, libgl_names[i] );
			while ( version_name ) {
				debug( "found version of \"%s\" as \"%s\", linking it\n", 
					   libgl_names[i], version_name );
				make_temp_link( tmpdir, version_name, wiregl_lib );
				version_name = find_next_version_name( dir, libgl_names[i] );
			}
			closedir( dir );
		} else {
			sys_error( "opendir( \"%s\" )", SYSTEM_LIB_DIR );
		}
	}

    prefix_env_var( tmpdir, "LD_LIBRARY_PATH" );

#if defined(IRIX) || defined(IRIX64)
	/* if these are set then they'll be used by rld, so we have to
       make sure to modify them. */
	if ( getenv( "LD_LIBRARYN32_PATH" ) )
		prefix_env_var( tmpdir, "LD_LIBRARYN32_PATH" );

	if ( getenv( "LD_LIBRARY64_PATH" ) )
		prefix_env_var( tmpdir, "LD_LIBRARY64_PATH" );
#endif

	pid = fork( );
	if ( pid < 0 )
		sys_fatal( "fork" );
	
	if ( pid == 0 ) {
		execvp( argv[0], argv );
		sys_fatal( "execvp \"%s\"", argv[0] );
	}

	debug( "started \"%s\" [%d]\n", argv[0], pid );

	do {
		while ( wait( &status ) != pid )
			;
	} while ( WIFSTOPPED(status) );

	delete_temp_files( );
	delete_temp_dirs( );

	if ( WIFEXITED(status) ) {
		debug( "\"%s\" exited, status=%d\n", argv[0], WEXITSTATUS(status) );
		exit( WEXITSTATUS(status) );
	}
		
	if ( WIFSIGNALED(status) ) {
		debug( "\"%s\" uncaught signal=%d\n", argv[0], WTERMSIG(status) );
		exit( 1 );
	}

	debug( "\"%s\" exited, not sure why\n", argv[0] );
	exit( 1 );
}

#endif

void
usage( void )
{
    char *wiregl_lib;

    fprintf( stderr,
             "usage: %s [options] <cmd> [args]\n"
             "options:\n"
             "   -lib    <file> : name of the replacement opengl32.dll\n"
			 "   -c[onf] <file> : name of configuration file\n"
             "   -v[erbose]\n"
             "   -q[uiet]\n"
             "   -h[elp]        : this message\n"
             "\n"
             "The environment variable WIREGL_LIB specifies the default -lib,\n",
             progname );

    wiregl_lib = getenv( "WIREGL_LIB" );
    if ( wiregl_lib )
        fprintf( stderr, "currently set to \"%s\".\n", wiregl_lib );
    else
        fprintf( stderr, "currently not set.\n" );
}

int
main( int argc, char **argv )
{
    int i;
	char *wiregl_conf;

    progname = basename( argv[0] );

    wiregl_lib = NULL;
	wiregl_conf = NULL;

    for ( i = 1; i < argc; i++ ) {

        if ( argv[i][0] != '-' )
            break;

        if ( !strcmp( argv[i], "-lib" ) ) {

            i++;
            if ( i == argc )
				fatal( "%s expects argument\n", argv[i-1] );

            wiregl_lib = argv[i];
        }
		else if ( !strcmp( argv[i], "-c" ) || !strcmp( argv[i], "-conf" ) ) {

			i++;
			if ( i == argc )
				fatal( "%s expects argument\n", argv[i-1] );

			wiregl_conf = argv[i];
		}
        else if ( !strcmp( argv[i], "-v" ) ||
                  !strcmp( argv[i], "-verbose" ) ) {
            verbose = 1;
        }
        else if ( !strcmp( argv[i], "-q" ) ||
                  !strcmp( argv[i], "-quiet" ) ) {
            verbose = 0;
        }
        else if ( !strcmp( argv[i], "-h" ) ||
                  !strcmp( argv[i], "-help" ) ) {

            usage( );
            exit( 0 );
        }
        else {

            error( "unknown argument: %s\n", argv[i] );
            usage( );
            exit( 1 );
        }
    }

    argc -= i;
    argv += i;

    if ( argc < 1 )
		fatal( "no command to run?\n" );

    if ( wiregl_lib == NULL ) {
        wiregl_lib = getenv( "WIREGL_LIB" );
        if ( wiregl_lib == NULL )
			fatal( "must specify -lib or WIREGL_LIB\n" );
    }

	if ( wiregl_conf )
		wgl_setenv( "WIREGL_CONF", wiregl_conf );

    do_it( argv );

    return 0;
}
